from __future__ import absolute_import
from ..errs import ValidationError
import json


class PrimJSONEncoder(json.JSONEncoder):
    """ json encoder for primitives
    """
    def default(self, obj):
        if hasattr(obj, 'to_json'):
            return obj.to_json()
        return json.JSONEncoder.default(self, obj)

def min_max(obj, val, is_max):
    """ min/max validator for float and integer
    """
    n = getattr(obj, 'maximum' if is_max else 'minimum', None)
    if n == None:
        return

    _eq = getattr(obj, 'exclusiveMaximum' if is_max else 'exclusiveMinimum', False)
    if is_max:
        to_raise = val >= n if _eq else val > n
    else:
        to_raise = val <= n if _eq else val < n

    if to_raise:
        raise ValidationError('condition failed: {0}, v:{1} compared to o:{2}'.format('maximum' if is_max else 'minimum', val, n))

#
# creater/2nd_pass function with python user defined class
# - Array
# - Model
# - Byte
# - Datetime
# - Date
# - File
#
def _2nd_pass_obj(obj, ret, val, ctx):
    return ret.apply_with(obj, val, ctx) 

def create_obj(obj, v, ctx=None, constructor=None):
    return constructor()




from __future__ import absolute_import
import six
from ..errs import ValidationError
from validate_email import validate_email


def validate_str(obj, ret, val, ctx):
    if obj.enum and ret not in obj.enum:
        raise ValidationError('{0} is not a valid enum for {1}'.format(ret, str(obj.enum)))
    if obj.maxLength and len(ret) > obj.maxLength:
        raise ValidationError('[{0}] is longer than {1} characters'.format(ret, str(obj.maxLength)))
    if obj.minLength and len(ret) < obj.minLength:
        raise ValidationError('[{0}] is shoter than {1} characters'.format(ret, str(obj.minLength)))

    # TODO: handle pattern
    return val

def create_str(obj, v, ctx=None):
    if isinstance(v, six.string_types):
        r = v
    else:
        r = str(v)
    validate_str(obj, r, v, ctx)
    return r

def validate_email_(obj, ret, val, ctx):
    if not validate_email(ret):
        raise ValidationError('{0} is not a valid email for {1}'.format(ret, obj))

    return val



from __future__ import absolute_import
from ..base import (
    Context,
    ContainerType,
    BaseObj,
    NullContext
    )
from .objects import (
    Schema,
    Swagger,
    Info,
    Items,
    Parameter,
    Header,
    Response,
    Operation,
    PathItem,
    SecurityScheme,
    Tag,
    ExternalDocumentation,
    Contact,
    License,
    XMLObject)


class ExternalDocumentationContext(Context):
    """ Context of External Document
    """

    __swagger_ref_object__ = ExternalDocumentation


class XMLObjectContext(Context):
    """ Context of XML Object
    """

    __swagger_ref_object__ = XMLObject


class ItemsContext(Context):
    """ Context of Item Object
    """

    __swagger_ref_object__ = Items

# self-reference
setattr(ItemsContext, '__swagger_child__', {
    'items': (None, ItemsContext),
})


class SchemaContext(Context):
    """ Context of Schema Object
    """

    __swagger_ref_object__ = Schema


class AdditionalPropertiesContext(Context):
    """ Context of additionalProperties,
    """

    class _TmpObj(BaseObj):
        def merge(self, other, _):
            if isinstance(other, bool):
                return other

            ret = Schema(NullContext())
            return ret.merge(other, SchemaContext)

    @classmethod
    def is_produced(kls, obj):
        """
        """
        if isinstance(obj, bool):
            return True
        return SchemaContext.is_produced(obj)

    def produce(self):
        """
        """
        if self.is_produced(self._obj):
            return self._obj
        else:
            return AdditionalPropertiesContext._TmpObj(self)

    def parse(self, obj=None):
        """
        """
        if obj == None:
            self._obj = True
        elif isinstance(obj, bool):
            self._obj = obj
        else:
            tmp = {'t': {}}
            with SchemaContext(tmp, 't') as ctx:
                ctx.parse(obj)

            self._obj = tmp['t']


# self-reference
setattr(SchemaContext, '__swagger_child__', {
    # items here should refer to an Schema Object.
    # refer to https://github.com/swagger-api/swagger-spec/issues/165
    # for details
    'items': (None, SchemaContext),
    'properties': (ContainerType.dict_, SchemaContext),
    # solution for properties with 2 possible types
    'additionalProperties': (None, AdditionalPropertiesContext),
    'allOf': (ContainerType.list_, SchemaContext),
    'xml': (None, XMLObjectContext),
    'externalDocs': (None, ExternalDocumentationContext),
})



class ParameterContext(Context):
    """ Context of Parameter Object, along with
    Reference Object
    """

    __swagger_child__ = {
        'schema': (None, SchemaContext),
        # items here should refer to an Items Object.
        # refer to https://github.com/swagger-api/swagger-spec/issues/165
        # for details
        'items': (None, ItemsContext),
    }
    __swagger_ref_object__ = Parameter


class HeaderContext(Context):
    """ Context of Header Object
    """

    __swagger_child__ = {
        'items': (None, ItemsContext),
    }
    __swagger_ref_object__ = Header


class ResponseContext(Context):
    """ Context of Response Object
    """

    __swagger_child__ = {
        'schema': (None, SchemaContext),
        'headers': (ContainerType.dict_, HeaderContext),
    }
    __swagger_ref_object__ = Response


class OperationContext(Context):
    """ Context of Operation Object
    """

    __swagger_child__ = {
        'parameters': (ContainerType.list_, ParameterContext),
        'responses': (ContainerType.dict_, ResponseContext),
    }
    __swagger_ref_object__ = Operation


class PathItemContext(Context):
    """ Context of Path Item Object
    """

    __swagger_child__ = {
        'get': (None, OperationContext),
        'put': (None, OperationContext),
        'post': (None, OperationContext),
        'delete': (None, OperationContext),
        'options': (None, OperationContext),
        'head': (None, OperationContext),
        'patch': (None, OperationContext),
        'parameters': (ContainerType.list_, ParameterContext),
    }
    __swagger_ref_object__ = PathItem


class SecuritySchemeContext(Context):
    """ Context of Security Schema Object
    """

    __swagger_ref_object__ = SecurityScheme



class TagContext(Context):
    """ Context of Tag Object
    """

    __swagger_ref_object__ = Tag
    __swagger_child__ = {
        'externalDocs': (None, ExternalDocumentationContext),
    }


class ContactContext(Context):
    """ Context of Contact Object
    """

    __swagger_ref_object__ = Contact


class LicenseContext(Context):
    """ Context of License Object
    """

    __swagger_ref_object__ = License


class InfoContext(Context):
    """ Context of Info Object
    """

    __swagger_ref_object__ = Info
    __swagger_child__ = {
        'contact': (None, ContactContext),
        'license': (None, LicenseContext),
    }


class SwaggerContext(Context):
    """ Context of Swagger Object
    """

    __swagger_child__ = {
        'info': (None, InfoContext),
        'paths': (ContainerType.dict_, PathItemContext),
        'definitions': (ContainerType.dict_, SchemaContext),
        'parameters': (ContainerType.dict_, ParameterContext),
        'responses': (ContainerType.dict_, ResponseContext),
        'securityDefinitions': (ContainerType.dict_, SecuritySchemeContext),
        'tags': (ContainerType.list_, TagContext),
        'externalDocs': (None, ExternalDocumentationContext),
    }
    __swagger_ref_object__ = Swagger



from __future__ import absolute_import
from .resolve import Resolver
from .primitives import Primitive, MimeCodec
from .spec.v1_2.parser import ResourceListContext
from .spec.v2_0.parser import SwaggerContext
from .spec.v2_0.objects import Operation
from .spec.base import BaseObj
from .scan import Scanner
from .scanner import TypeReduce, CycleDetector
from .scanner.v1_2 import Upgrade
from .scanner.v2_0 import AssignParent, Merge, Resolve, PatchObject, YamlFixer, Aggregate, NormalizeRef
from pyswagger import utils, errs, consts
import copy
import base64
import six
import weakref
import logging


logger = logging.getLogger(__name__)


class App(object):
    """ Major component of pyswagger

    This object is tended to be used in read-only manner. Therefore,
    all accessible attributes are almost read-only properties.
    """

    sc_path = 1

    _shortcut_ = {
        sc_path: ('/', '#/paths')
    }

    def __init__(self, url=None, url_load_hook=None, sep=consts.private.SCOPE_SEPARATOR, prim=None, mime_codec=None, resolver=None):
        """ constructor

        :param url str: url of swagger.json
        :param func url_load_hook: a way to redirect url to a accessible place. for self testing.
        :param sep str: separator used by pyswager.utils.ScopeDict
        :param prim pyswagger.primitives.Primitive: factory for primitives in Swagger.
        :param resolver: pyswagger.resolve.Resolver: customized resolver used as default when none is provided when resolving
        """

        logger.info('init with url: {0}'.format(url))

        self.__root = None
        self.__raw = None
        self.__version = ''

        self.__op = None
        self.__m = None
        self.__schemes = []
        self.__url=url

        # a map from json-reference to
        # - spec.BaseObj
        # - a map from json-pointer to spec.BaseObj
        self.__objs = {}

        if url_load_hook and resolver:
            raise ValueError('when use customized Resolver, please pass url_load_hook to that one')

        # the start-point when you want to traverse the code to laod new object
        self.__resolver = resolver or Resolver(url_load_hook)

        # allow init App-wised SCOPE_SEPARATOR
        self.__sep = sep

        # init a default Primitive as factory for primitives
        self.__prim = prim if prim else Primitive()

        # MIME codec
        self.__mime_codec = mime_codec or MimeCodec()

    @property
    def root(self):
        """ schema representation of Swagger API, its structure may
        be different from different version of Swagger.

        There is 'Schema' object in swagger 2.0, that's why I change this
        property name from 'schema' to 'root'.

        :type: pyswagger.spec.v2_0.objects.Swagger
        """
        return self.__root

    @property
    def raw(self):
        """ raw objects for original version of loaded resources.
        When loaded json is the latest version we supported, this property is the same as App.root

        :type: ex. when loading Swagger 1.2, the type is pyswagger.spec.v1_2.objects.ResourceList
        """
        return self.__raw

    @property
    def op(self):
        """ list of Operations, organized by utils.ScopeDict

        In Swagger 2.0, Operation(s) can be organized with Tags and Operation.operationId.
        ex. if there is an operation with tag:['user', 'security'] and operationId:get_one,
        here is the combination of keys to access it:
        - .op['user', 'get_one']
        - .op['security', 'get_one']
        - .op['get_one']

        :type: pyswagger.utils.ScopeDict of pyswagger.spec.v2_0.objects.Operation
        """
        return self.__op

    @property
    def m(self):
        """ backward compatible to access Swagger.definitions in Swagger 2.0,
        and Resource.Model in Swagger 1.2.

        ex. a Model:user in Resource:Users, access it by .m['Users', 'user'].
        For Schema object in Swagger 2.0, just access it by it key in json.

        :type: pyswagger.utils.ScopeDict
        """
        return self.__m

    @property
    def version(self):
        """ original version of loaded json

        :type: str
        """
        return self.__version

    @property
    def schemes(self):
        """ supported schemes, refer to Swagger.schemes in Swagger 2.0 for details

        :type: list of str, ex. ['http', 'https']
        """
        return self.__schemes

    @property
    def url(self):
        """
        """
        return self.__url

    @property
    def prim_factory(self):
        """ primitive factory used by this app

        :type: pyswagger.primitives.Primitive
        """
        return self.__prim

    @property
    def mime_codec(self):
        """ mime codec used by this app

        :type: pyswagger.primitives.MimeCodec
        """
        return self.__mime_codec

    def load_obj(self, jref, getter=None, parser=None):
        """ load a object(those in spec._version_.objects) from a JSON reference.
        """
        obj = self.__resolver.resolve(jref, getter)

        # get root document to check its swagger version.
        tmp = {'_tmp_': {}}
        version = utils.get_swagger_version(obj)
        if version == '1.2':
            # swagger 1.2
            with ResourceListContext(tmp, '_tmp_') as ctx:
                ctx.parse(obj, jref, self.__resolver, getter)
        elif version == '2.0':
            # swagger 2.0
            with SwaggerContext(tmp, '_tmp_') as ctx:
                ctx.parse(obj)
        elif version == None and parser:
            with parser(tmp, '_tmp_') as ctx:
                ctx.parse(obj)

            version = tmp['_tmp_'].__swagger_version__ if hasattr(tmp['_tmp_'], '__swagger_version__') else version
        else:
            raise NotImplementedError('Unsupported Swagger Version: {0} from {1}'.format(version, jref))

        if not tmp['_tmp_']:
            raise Exception('Unable to parse object from {0}'.format(jref))

        logger.info('version: {0}'.format(version))

        return tmp['_tmp_'], version

    def prepare_obj(self, obj, jref):
        """ basic preparation of an object(those in sepc._version_.objects),
        and cache the 'prepared' object.
        """
        if not obj:
            raise Exception('unexpected, passing {0}:{1} to prepare'.format(obj, jref))

        s = Scanner(self)
        if self.version == '1.2':
            # upgrade from 1.2 to 2.0
            converter = Upgrade(self.__sep)
            s.scan(root=obj, route=[converter])
            obj = converter.swagger

            if not obj:
                raise Exception('unable to upgrade from 1.2: {0}'.format(jref))

            s.scan(root=obj, route=[AssignParent()])

        # normalize $ref
        url, jp = utils.jr_split(jref)
        s.scan(root=obj, route=[NormalizeRef(url)])
        # fix for yaml that treat response code as number
        s.scan(root=obj, route=[YamlFixer()], leaves=[Operation])

        # cache this object
        if url not in self.__objs:
            if jp == '#':
                self.__objs[url] = obj
            else:
                self.__objs[url] = {jp: obj}
        else:
            if not isinstance(self.__objs[url], dict):
                raise Exception('it should be able to resolve with BaseObj')
            self.__objs[url].update({jp: obj})

        # pre resolve Schema Object
        # note: make sure this object is cached before using 'Resolve' scanner
        s.scan(root=obj, route=[Resolve()])
        return obj

    def _validate(self):
        """ check if this Swagger API valid or not.

        :param bool strict: when in strict mode, exception would be raised if not valid.
        :return: validation errors
        :rtype: list of tuple(where, type, msg).
        """

        v_mod = utils.import_string('.'.join([
            'pyswagger',
            'scanner',
            'v' + self.version.replace('.', '_'),
            'validate'
        ]))

        if not v_mod:
            # there is no validation module
            # for this version of spec
            return []

        s = Scanner(self)
        v = v_mod.Validate()

        s.scan(route=[v], root=self.__raw)
        return v.errs

    @classmethod
    def load(kls, url, getter=None, parser=None, url_load_hook=None, sep=consts.private.SCOPE_SEPARATOR, prim=None, mime_codec=None, resolver=None):
        """ load json as a raw App

        :param str url: url of path of Swagger API definition
        :param getter: customized Getter
        :type getter: sub class/instance of Getter
        :param parser: the parser to parse the loaded json.
        :type parser: pyswagger.base.Context
        :param dict app_cache: the cache shared by related App
        :param func url_load_hook: hook to patch the url to load json
        :param str sep: scope-separater used in this App
        :param prim pyswager.primitives.Primitive: factory for primitives in Swagger
        :param mime_codec pyswagger.primitives.MimeCodec: MIME codec
        :param resolver: pyswagger.resolve.Resolver: customized resolver used as default when none is provided when resolving
        :return: the created App object
        :rtype: App
        :raises ValueError: if url is wrong
        :raises NotImplementedError: the swagger version is not supported.
        """

        logger.info('load with [{0}]'.format(url))

        url = utils.normalize_url(url)
        app = kls(url, url_load_hook=url_load_hook, sep=sep, prim=prim, mime_codec=mime_codec, resolver=resolver)
        app.__raw, app.__version = app.load_obj(url, getter=getter, parser=parser)
        if app.__version not in ['1.2', '2.0']:
            raise NotImplementedError('Unsupported Version: {0}'.format(self.__version))

        # update schem if any
        p = six.moves.urllib.parse.urlparse(url)
        if p.scheme:
            app.schemes.append(p.scheme)

        return app

    def validate(self, strict=True):
        """ check if this Swagger API valid or not.

        :param bool strict: when in strict mode, exception would be raised if not valid.
        :return: validation errors
        :rtype: list of tuple(where, type, msg).
        """

        result = self._validate()
        if strict and len(result):
            for r in result:
                logger.error(r)
            raise errs.ValidationError('this Swagger App contains error: {0}.'.format(len(result)))

        return result

    def prepare(self, strict=True):
        """ preparation for loaded json

        :param bool strict: when in strict mode, exception would be raised if not valid.
        """

        self.validate(strict=strict)
        self.__root = self.prepare_obj(self.raw, self.__url)

        if hasattr(self.__root, 'schemes') and self.__root.schemes:
            if len(self.__root.schemes) > 0:
                self.__schemes = self.__root.schemes
            else:
                # extract schemes from the url to load spec
                self.__schemes = [six.moves.urlparse(self.__url).schemes]

        s = Scanner(self)
        s.scan(root=self.__root, route=[Merge()])
        s.scan(root=self.__root, route=[PatchObject()])
        s.scan(root=self.__root, route=[Aggregate()])

        # reducer for Operation
        tr = TypeReduce(self.__sep)
        cy = CycleDetector()
        s.scan(root=self.__root, route=[tr, cy])

        # 'op' -- shortcut for Operation with tag and operaionId
        self.__op = utils.ScopeDict(tr.op)
        # 'm' -- shortcut for model in Swagger 1.2
        if hasattr(self.__root, 'definitions') and self.__root.definitions != None:
            self.__m = utils.ScopeDict(self.__root.definitions)
        else:
            self.__m = utils.ScopeDict({})
        # update scope-separater
        self.__m.sep = self.__sep
        self.__op.sep = self.__sep

        # cycle detection
        if len(cy.cycles['schema']) > 0 and strict:
            raise errs.CycleDetectionError('Cycles detected in Schema Object: {0}'.format(cy.cycles['schema']))

    @classmethod
    def create(kls, url, strict=True):
        """ factory of App

        :param str url: url of path of Swagger API definition
        :param bool strict: when in strict mode, exception would be raised if not valid.
        :return: the created App object
        :rtype: App
        :raises ValueError: if url is wrong
        :raises NotImplementedError: the swagger version is not supported.
        """
        app = kls.load(url)
        app.prepare(strict=strict)

        return app

    """ for backward compatible, for later version,
    please call App.create instead.
    """
    _create_ = create

    def resolve(self, jref, parser=None):
        """ JSON reference resolver

        :param str jref: a JSON Reference, refer to http://tools.ietf.org/html/draft-pbryan-zyp-json-ref-03 for details.
        :param parser: the parser corresponding to target object.
        :type parser: pyswagger.base.Context
        :return: the referenced object, wrapped by weakref.ProxyType
        :rtype: weakref.ProxyType
        :raises ValueError: if path is not valid
        """

        logger.info('resolving: [{0}]'.format(jref))

        if jref == None or len(jref) == 0:
            raise ValueError('Empty Path is not allowed')

        obj = None
        url, jp = utils.jr_split(jref)

        # check cacahed object against json reference by
        # comparing url first, and find those object prefixed with
        # the JSON pointer.
        o = self.__objs.get(url, None)
        if o:
            if isinstance(o, BaseObj):
                obj = o.resolve(utils.jp_split(jp)[1:])
            elif isinstance(o, dict):
                for k, v in six.iteritems(o):
                    if jp.startswith(k):
                        obj = v.resolve(utils.jp_split(jp[len(k):])[1:])
                        break
            else:
                raise Exception('Unknown Cached Object: {0}'.format(str(type(o))))

        # this object is not found in cache
        if obj == None:
            if url:
                obj, _ = self.load_obj(jref, parser=parser)
                if obj:
                    obj = self.prepare_obj(obj, jref)
            else:
                # a local reference, 'jref' is just a json-pointer
                if not jp.startswith('#'):
                    raise ValueError('Invalid Path, root element should be \'#\', but [{0}]'.format(jref))

                obj = self.root.resolve(utils.jp_split(jp)[1:]) # heading element is #, mapping to self.root

        if obj == None:
            raise ValueError('Unable to resolve path, [{0}]'.format(jref))

        if isinstance(obj, (six.string_types, six.integer_types, list, dict)):
            return obj
        return weakref.proxy(obj)

    def s(self, p, b=_shortcut_[sc_path]):
        """ shortcut of App.resolve.
        We provide a default base for '#/paths'. ex. to access '#/paths/~1user/get',
        just call App.s('user/get')

        :param str p: path relative to base
        :param tuple b: a tuple (expected_prefix, base) to represent a 'base'
        """

        if b[0]:
            return self.resolve(utils.jp_compose(b[0] + p if not p.startswith(b[0]) else p, base=b[1]))
        else:
            return self.resolve(utils.jp_compose(p, base=b[1]))

    def dump(self):
        """ dump into Swagger Document

        :rtype: dict
        :return: dict representation of Swagger
        """
        return self.root.dump()



class Security(object):
    """ security handler
    """

    def __init__(self, app):
        """ constructor

        :param App app: App
        """
        self.__app = app

        # placeholder of Security Info
        self.__info = {}

    def update_with(self, name, security_info):
        """ insert/clear authorizations

        :param str name: name of the security info to be updated
        :param security_info: the real security data, token, ...etc.
        :type security_info: **(username, password)** for *basicAuth*, **token** in str for *oauth2*, *apiKey*.

        :raises ValueError: unsupported types of authorizations
        """
        s = self.__app.root.securityDefinitions.get(name, None)
        if s == None:
            raise ValueError('Unknown security name: [{0}]'.format(name))

        cred = security_info
        header = True
        if s.type == 'basic':
            cred = 'Basic ' + base64.standard_b64encode(six.b('{0}:{1}'.format(*security_info))).decode('utf-8')
            key = 'Authorization'
        elif s.type == 'apiKey':
            key = s.name
            header = getattr(s, 'in') == 'header'
        elif s.type == 'oauth2':
            key = 'access_token'
        else:
            raise ValueError('Unsupported Authorization type: [{0}, {1}]'.format(name, s.type))

        self.__info.update({name: (header, {key: cred})})

    def __call__(self, req):
        """ apply security info for a request.

        :param Request req: the request to be authorized.
        :return: the updated request
        :rtype: Request
        """
        if not req._security:
            return req

        for s in req._security:
            for k, v in six.iteritems(s):
                if not k in self.__info:
                    logger.info('missing: [{0}]'.format(k))
                    continue

                logger.info('applying: [{0}]'.format(k))

                header, cred = self.__info[k]
                if header:
                    req._p['header'].update(cred)
                else:
                    utils.nv_tuple_list_replace(req._p['query'], utils.get_dict_as_tuple(cred))

        return req


class BaseClient(object):
    """ base implementation of SwaggerClient, below is an minimum example
    to extend this class

    .. code-block:: python

        class MyClient(BaseClient):

            # declare supported schemes here
            __schemes__ = ['http', 'https']

            def request(self, req_and_resp, opt):
                # passing to parent for default patching behavior,
                # applying authorizations, ...etc.
                req, resp = super(MyClient, self).request(req_and_resp, opt)

                # perform request by req
                ...
                # apply result to resp
                resp.apply(header=header, raw=data_received, status=code)
                return resp
    """

    join_headers = 'join_headers'

    # supported schemes, ex. ['http', 'https', 'ws', 'ftp']
    __schemes__ = set()

    def __init__(self, security=None):
        """ constructor

        :param Security security: the security holder
        """

        # placeholder of Security
        self.__security = security

    def prepare_schemes(self, req):
        """ make sure this client support schemes required by current request

        :param pyswagger.io.Request req: current request object
        """

        # fix test bug when in python3 scheme, more details in commint msg
        ret = sorted(self.__schemes__ & set(req.schemes), reverse=True)

        if len(ret) == 0:
            raise ValueError('No schemes available: {0}'.format(req.schemes))
        return ret

    def compose_headers(self, req, headers=None, opt=None, as_dict=False):
        """ a utility to compose headers from pyswagger.io.Request and customized headers

        :return: list of tuple (key, value) when as_dict is False, else dict
        """
        if headers is None:
            return list(req.header.items()) if not as_dict else req.header

        opt = opt or {}
        join_headers = opt.pop(BaseClient.join_headers, None)
        if as_dict and not join_headers:
            # pick the most efficient way for special case
            headers = dict(headers) if isinstance(headers, list) else headers
            headers.update(req.header)
            return headers

        # include Request.headers
        aggregated_headers = list(req.header.items())

        # include customized headers
        if isinstance(headers, list):
            aggregated_headers.extend(headers)
        elif isinstance(headers, dict):
            aggregated_headers.extend(headers.items())
        else:
            raise Exception('unknown type as header: {}'.format(str(type(headers))))

        if join_headers:
            joined = {}
            for h in aggregated_headers:
                key = h[0]
                if key in joined:
                    joined[key] = ','.join([joined[key], h[1]])
                else:
                    joined[key] = h[1]
            aggregated_headers = list(joined.items())

        return dict(aggregated_headers) if as_dict else aggregated_headers

    def request(self, req_and_resp, opt=None, headers=None):
        """ preprocess before performing a request, usually some patching.
        authorization also applied here.

        :param req_and_resp: tuple of Request and Response
        :type req_and_resp: (Request, Response)
        :param opt: customized options
        :type opt: dict
        :param headers: customized headers
        :type headers: dict of 'string', or list of tuple: (string, string) for multiple values for one key
        :return: patched request and response
        :rtype: Request, Response
        """
        req, resp = req_and_resp

        # dump info for debugging
        logger.info('request.url: {0}'.format(req.url))
        logger.info('request.header: {0}'.format(req.header))
        logger.info('request.query: {0}'.format(req.query))
        logger.info('request.file: {0}'.format(req.files))
        logger.info('request.schemes: {0}'.format(req.schemes))

        # apply authorizations
        if self.__security:
            self.__security(req)

        return req, resp


SwaggerApp = App
SwaggerSecurity = Security




from __future__ import absolute_import
from .spec.base import BaseObj
import six


def default_tree_traversal(root, leaves):
    """ default tree traversal """
    objs = [('#', root)]
    while len(objs) > 0:
        path, obj = objs.pop()

        # name of child are json-pointer encoded, we don't have
        # to encode it again.
        if obj.__class__ not in leaves:
            objs.extend(map(lambda i: (path + '/' + i[0],) + (i[1],), six.iteritems(obj._children_)))

        # the path we expose here follows JsonPointer described here
        #   http://tools.ietf.org/html/draft-ietf-appsawg-json-pointer-07
        yield path, obj


class DispatcherMeta(type):
    """ metaclass for Dispatcher
    """
    def __new__(metacls, name, bases, spc):
        if 'obj_route' not in spc.keys():
            # forcely create a new obj_route
            # but not share the same one with parents.
            spc['obj_route'] = {}
            spc['result_fn'] = [None]

        return type.__new__(metacls, name, bases, spc)


class Dispatcher(six.with_metaclass(DispatcherMeta, object)):
    """ Dispatcher
    """
    obj_route = {}
    result_fn = [None]

    @classmethod
    def __add_route(cls, t, f):
        """
        """
        if not issubclass(t, BaseObj):
            raise ValueError('target_cls should be a subclass of BaseObj, but got:' + str(t))

        # allow register multiple handler function
        # against one object
        if t in cls.obj_route.keys():
            cls.obj_route[t].append(f)
        else:
            cls.obj_route[t] = [f]

    @classmethod
    def register(cls, target):
        """
        """
        def outer_fn(f):
            # what we did is simple,
            # register target_cls as key, and f as callback
            # then keep this record in cls.
            for t in target:
                cls.__add_route(t, f)

            # nothing is decorated. Just return original one.
            return f

        return outer_fn

    @classmethod
    def result(cls, f):
        """
        """

        # avoid bound error
        cls.result_fn = [f]
        return f


class Scanner(object):
    """ Scanner
    """
    def __init__(self, app):
        super(Scanner, self).__init__()
        self.__app = app

    @property
    def app(self):
        return self.__app

    def __build_route(self, route):
        """
        """
        ret = []
        for r in route:
            for attr in r.__class__.__dict__:
                o = getattr(r, attr)
                if type(o) == DispatcherMeta:
                    ret.append((r, o.obj_route, o.result_fn[0]))

        return ret

    def scan(self, route, root, nexter=default_tree_traversal, leaves=[]):
        """
        """
        if root == None:
            raise ValueError('Can\'t scan because root==None')

        merged_r = self.__build_route(route)
        for path, obj in nexter(root, leaves):
            for the_self, r, res in merged_r:

                def handle_cls(cls):
                    f = r.get(cls, None)
                    if f:
                        for ff in f:
                            ret = ff(the_self, path, obj, self.app)
                            if res:
                                res(the_self, ret)

                for cls in obj.__class__.__mro__[:-1]:
                    if cls is BaseObj:
                        break
                    handle_cls(cls)




from __future__ import absolute_import
from ..errs import ValidationError


class File(object):
    """ for type File
    """
    def apply_with(self, _, val, ctx):
        """ constructor

        example val:
        {
            # header values used in multipart/form-data according to RFC2388
            'header': {
                'Content-Type': 'text/plain',
                
                # according to RFC2388, available values are '7bit', '8bit', 'binary'
                'Content-Transfer-Encoding': 'binary'
            },
            filename: 'a.txt',
            data: None (or any file-like object)
        }

        :param val: dict containing file info.
        """
        self.header = val.get('header', {})
        self.data = val.get('data', None)
        self.filename = val.get('filename', '')
        if self.data == None and self.filename == '':
            raise ValidationError('should have file name or file object, not: {0}, {1}'.format(
                self.data, self.filename
            ))



from __future__ import absolute_import
from ...scan import Dispatcher
from ...spec.v1_2.objects import (
    DataTypeObj,
    Parameter,
    Property,
    Items,
    Authorization,
    Operation,
    GrantType,
    Scope,
    Authorizations,
    LoginEndpoint,
    Implicit,
    TokenRequestEndpoint,
    TokenEndpoint,
    AuthorizationCode,
    ResponseMessage,
    Api,
    Model,
    Resource,
    Info,
    ResourceList
    )
import six


class Validate(object):
    """
    """
    class Disp(Dispatcher): pass

    def __init__(self):
        self.errs = []

    @Disp.register([DataTypeObj])
    def _validate_data(self, path, obj, _):
        """ convert string value to integer/float value """
        errs = []

        if obj.type == None and getattr(obj, '$ref') == None:
            errs.append('type or $ref should be existed.')

        # make sure there is 'item' along with 'array'
        if not (obj.type == 'array') == (obj.items != None):
            errs.append('array should be existed along with items')
        if obj.uniqueItems != None and obj.type != 'array':
            errs.append('uniqueItems is only used for array type.')

        return path, obj.__class__.__name__, errs

    @Disp.register([Parameter])
    def _validate_param(self, path, obj, _):
        """ validate option combination of Parameter object """
        errs = []

        if obj.allowMultiple:
            if not obj.paramType in ('path', 'query', 'header'):
                errs.append('allowMultiple should be applied on path, header, or query parameters')
            if obj.type == 'array':
                errs.append('array Type with allowMultiple is not supported.')

        if obj.paramType == 'body' and obj.name not in ('', 'body'):
            errs.append('body parameter with invalid name: {0}'.format(obj.name))

        if obj.type == 'File':
            if obj.paramType != 'form':
                errs.append('File parameter should be form type: {0}'.format(obj.name))
            if 'multipart/form-data' not in obj._parent_.consumes:
                errs.append('File parameter should consume multipart/form-data: {0}'.format(obj.name))

        if obj.type == 'void':
            errs.append('void is only allowed in Operation object.')

        return path, obj.__class__.__name__, errs

    @Disp.register([Property])
    def _validate_prop(self, path, obj, _):
        """ validate option combination of Property object """
        errs = []

        if obj.type == 'void':
            errs.append('void is only allowed in Operation object.')

        return path, obj.__class__.__name__, errs

    @Disp.register([Items])
    def _validate_items(self, path, obj, _):
        """ validate option combination of Property object """
        errs = []

        if obj.type == 'void':
            errs.append('void is only allowed in Operation object.')

        return path, obj.__class__.__name__, errs

    @Disp.register([Authorization])
    def _validate_auth(self, path, obj, _):
        """ validate that apiKey and oauth2 requirements """
        errs = []

        if obj.type == 'apiKey':
            if not obj.passAs:
                errs.append('need "passAs" for apiKey')
            if not obj.keyname:
                errs.append('need "keyname" for apiKey')

        elif obj.type == 'oauth2':
            if not obj.grantTypes:
                errs.append('need "grantTypes" for oauth2')

        return path, obj.__class__.__name__, errs

    @Disp.register([GrantType])
    def _validate_granttype(self, path, obj, _):
        """ make sure either implicit or authorization_code is defined """
        errs = []

        if not obj.implicit and not obj.authorization_code:
            errs.append('Either implicit or authorization_code should be defined.')

        return path, obj.__class__.__name__, errs

    @Disp.register([Operation])
    def _validate_auths(self, path, obj, app):
        """ make sure that apiKey and basicAuth are empty list
        in Operation object.
        """
        errs = []

        for k, v in six.iteritems(obj.authorizations or {}):
            if k not in app.raw.authorizations:
                errs.append('auth {0} not found in resource list'.format(k))

            if app.raw.authorizations[k].type in ('basicAuth', 'apiKey') and v != []:
                errs.append('auth {0} should be an empty list'.format(k))

        return path, obj.__class__.__name__, errs

    """ requirement """

    @staticmethod
    def _check_reqs(path, obj, reqs):
        errs = []
        for r in reqs:
           if not hasattr(obj, r) or None == getattr(obj, r):
                errs.append('requirement {0} not meet.'.format(r))

        return path, obj.__class__.__name__, errs

    @Disp.register([Scope])
    def scope_require(self, path, obj, _):
        return self._check_reqs(path, obj, ['scope'])

    @Disp.register([Authorizations])
    def authorizations_require(self, path, obj, _):
        return self._check_reqs(path, obj, ['scope'])

    @Disp.register([LoginEndpoint])
    def login_endpoint_require(self, path, obj, _):
        return self._check_reqs(path, obj, ['url'])

    @Disp.register([Implicit])
    def implicit_require(self, path, obj, _):
        return self._check_reqs(path, obj, ['loginEndpoint'])

    @Disp.register([TokenRequestEndpoint])
    def token_request_endpoint_require(self, path, obj, _):
        return self._check_reqs(path, obj, ['url'])

    @Disp.register([TokenEndpoint])
    def token_endpoint_require(self, path, obj, _):
        return self._check_reqs(path, obj, ['url'])

    @Disp.register([AuthorizationCode])
    def authorization_code_require(self, path, obj, _):
        return self._check_reqs(path, obj, ['tokenRequestEndpoint', 'tokenEndpoint'])


    @Disp.register([Authorization])
    def authorization_require(self, path, obj, _):
        return self._check_reqs(path, obj, ['type'])

    @Disp.register([ResponseMessage])
    def response_message_require(self, path, obj, _):
        return self._check_reqs(path, obj, ['code', 'message'])

    @Disp.register([Parameter])
    def parameter_require(self, path, obj, _):
        return self._check_reqs(path, obj, ['paramType', 'name'])

    @Disp.register([Operation])
    def operation_require(self, path, obj, _):
        return self._check_reqs(path, obj, ['method', 'nickname', 'parameters'])

    @Disp.register([Api])
    def api_require(self, path, obj, _):
        return self._check_reqs(path, obj, ['path', 'operations'])

    @Disp.register([Model])
    def model_require(self, path, obj, _):
        return self._check_reqs(path, obj, ['id', 'properties'])

    @Disp.register([Resource])
    def resource_require(self, path, obj, _):
        return self._check_reqs(path, obj, ['swaggerVersion', 'basePath', 'apis'])

    @Disp.register([ResourceList])
    def resource_list_require(self, path, obj, _):
        return self._check_reqs(path, obj, ['swaggerVersion', 'apis'])

    @Disp.register([Info])
    def info_require(self, path, obj, _):
        return self._check_reqs(path, obj, ['title', 'description'])

    @Disp.result
    def result(self, result):
        """ aggregate result """
        if result:
            if len(result[2]) == 0:
                return
        else:
            return

        self.errs.extend([((result[0], result[1]), err) for err in result[2]])




from __future__ import absolute_import
from ..base import BaseObj, FieldMeta, Context
import six
import copy


class BaseObj_v1_2(BaseObj):
    __swagger_version__ = '1.2'


class Items(six.with_metaclass(FieldMeta, BaseObj_v1_2)):
    """ Items Object
    """
    __swagger_fields__ = {
        '$ref': None,
        'type': None,
        'format': None,
    }


class ItemsContext(Context):
    """ Context of Items Object
    """
    __swagger_ref_object__ = Items


class DataTypeObj(BaseObj_v1_2):
    """ Data Type Fields
    """
    __swagger_fields__ = {
        'type': None,
        '$ref': None,
        'format': None,
        'defaultValue': None,
        'enum': None,
        'items': None,
        'minimum': None,
        'maximum': None,
        'uniqueItems': None,
    }

    def __init__(self, ctx):
        # Items Object, too lazy to create a Context for DataTypeObj
        # to wrap this child.
        items_data = ctx._obj.get('items', None)
        if items_data:
            with ItemsContext(ctx._obj, 'items') as items_ctx:
                items_ctx.parse(items_data)
        else:
            setattr(self, self.get_private_name('items'), None)

        super(DataTypeObj, self).__init__(ctx)

class Scope(six.with_metaclass(FieldMeta, BaseObj_v1_2)):
    """ Scope Object
    """

    __swagger_fields__ = {
        'scope': None,
        'description': None,
    }


class LoginEndpoint(six.with_metaclass(FieldMeta, BaseObj_v1_2)):
    """ LoginEndpoint Object
    """

    __swagger_fields__ = {
        'url': None,
    }


class Implicit(six.with_metaclass(FieldMeta, BaseObj_v1_2)):
    """ Implicit Object
    """

    __swagger_fields__ = {
        'loginEndpoint': None,
        'tokenName': None,
    }


class TokenRequestEndpoint(six.with_metaclass(FieldMeta, BaseObj_v1_2)):
    """ TokenRequestEndpoint Object
    """

    __swagger_fields__ = {
        'url': None,
        'clientIdName': None,
        'clientSecretName': None,
    }


class TokenEndpoint(six.with_metaclass(FieldMeta, BaseObj_v1_2)):
    """ TokenEndpoint Object
    """

    __swagger_fields__ = {
        'url': None,
        'tokenName': None,
    }


class AuthorizationCode(six.with_metaclass(FieldMeta, BaseObj_v1_2)):
    """ AuthorizationCode Object
    """

    __swagger_fields__ = {
        'tokenRequestEndpoint': None,
        'tokenEndpoint': None,
    }


class GrantType(six.with_metaclass(FieldMeta, BaseObj_v1_2)):
    """ GrantType Object
    """

    __swagger_fields__ = {
        'implicit': None,
        'authorization_code': None,
    }


class Authorizations(six.with_metaclass(FieldMeta, BaseObj_v1_2)):
    """ Authorizations Object
    """

    __swagger_fields__ = {
        'scope': None,
        'description': None,
    }


class Authorization(six.with_metaclass(FieldMeta, BaseObj_v1_2)):
    """ Authorization Object
    """

    __swagger_fields__ = {
        'type': None,
        'passAs': None,
        'keyname': None,
        'scopes': None,
        'grantTypes': None,
    }

    def get_name(self, path):
        return path.split('/', 3)[2]


class ResponseMessage(six.with_metaclass(FieldMeta, BaseObj_v1_2)):
    """ ResponseMessage Object
    """

    __swagger_fields__ = {
        'code': None,
        'message': None,
        'responseModel': None,
    }


class Parameter(six.with_metaclass(FieldMeta, DataTypeObj)):
    """ Parameter Object
    """

    __swagger_fields__ = {
        'paramType': None,
        'name': None,
        'required': None,
        'allowMultiple': None,
        'description': None,
    }


class Operation(six.with_metaclass(FieldMeta, DataTypeObj)):
    """ Operation Object
    """

    __swagger_fields__ = {
        'method': None,
        'nickname': None,
        'authorizations': None,
        'parameters': None,
        'responseMessages': None,
        'produces': None,
        'consumes': None,
        'deprecated': None,
        'summary': None,
        'notes': None,
    }

    __internal_fields__ = {
        # path from Api object, concated with Resource object
        'path': None,
    }

    def get_name(self, path):
        return self.nickname


class Api(six.with_metaclass(FieldMeta, BaseObj_v1_2)):
    """ Api Object
    """

    __swagger_fields__ = {
        'path': None,
        'operations': None,
        'description': None,
    }


class Property(six.with_metaclass(FieldMeta, DataTypeObj)):
    """ Property Object
    """

    __swagger_fields__ = {
        'description': None,
    }     


class Model(six.with_metaclass(FieldMeta, BaseObj_v1_2)):
    """ Model Object
    """

    __swagger_fields__ = {
        'id': None,
        'required': [],
        'properties': None,
        'subTypes': None,
        'discriminator': None,
        'description': None,
    }

    __internal_fields__ = {
        # for model inheritance
        '_extends_': None,
    }

    def get_name(self, path):
        return self.id


class Resource(six.with_metaclass(FieldMeta, BaseObj_v1_2)):
    """ Resource Object
    """

    __swagger_fields__ = {
        'swaggerVersion': None,
        'apiVersion': None,
        'apis': None,
        'basePath': None,
        'resourcePath': None,
        'models': None,
        'produces': None,
        'consumes': None,
        'authorizations': None,
        'description': None,
    }

    def __init__(self, ctx):
        """ The original structure of API object is very bad
        for seeking nickname for operations. Since nickname is unique
        in one Resource, we can just make it flat.
        """
        super(Resource, self).__init__(ctx)

        new_api = {}
        for api in ctx._obj['apis']:
            for op in api.operations:
                name = op.nickname
                if name in new_api.keys():
                    raise ValueError('duplication operation found: ' + name)

                # Operation objects now have 'path' attribute.
                op.update_field('path', api.path)
                # Operation objects' parent is now Resource object(API Declaration).
                op._parent__ = self
                new_api[name] = op

        # replace Api with Operations
        self.update_field('apis', new_api)

    def get_name(self, path):
        return path.split('/', 3)[2]


class Info(six.with_metaclass(FieldMeta, BaseObj_v1_2)):
    """ Info Object
    """

    __swagger_fields__ = {
        'title': None,
        'termsOfServiceUrl': None,
        'contact': None,
        'license': None,
        'licenseUrl': None,
        'description': None,
    }


class ResourceList(six.with_metaclass(FieldMeta, BaseObj_v1_2)):
    """ Resource List Object
    """
    __swagger_fields__ = {
        'swaggerVersion': None,
        'apis': None,
        'apiVersion': None,
        'info': None,
        'authorizations': None,
    }




from __future__ import absolute_import
import six

class Model(dict):
    """ for complex type: models
    """

    # access dict like object
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__

    def __init__(self):
        """ constructor
        """
        super(Model, self).__init__()

    def apply_with(self, obj, val, ctx):
        """ recursivly apply Schema object

        :param obj.Model obj: model object to instruct how to create this model
        :param dict val: things used to construct this model
        """
        for k, v in six.iteritems(val):
            if k in obj.properties:
                pobj = obj.properties.get(k)
                if pobj.readOnly == True and ctx['read'] == False:
                    raise Exception('read-only property is set in write context.')

                self[k] = ctx['factory'].produce(pobj, v)

            # TODO: patternProperties here
            # TODO: fix bug, everything would not go into additionalProperties, instead of recursive
            elif obj.additionalProperties == True:
                ctx['addp'] = True
            elif obj.additionalProperties not in (None, False):
                ctx['addp_schema'] = obj

        in_obj = set(six.iterkeys(obj.properties))
        in_self = set(six.iterkeys(self))

        other_prop = in_obj - in_self
        for k in other_prop:
            p = obj.properties[k]
            if p.is_set("default"):
                self[k] = ctx['factory'].produce(p, p.default)

        not_found = set(obj.required) - set(six.iterkeys(self))
        if len(not_found):
            raise ValueError('requirement not meet: {0}'.format(not_found))

        # remove assigned properties to avoid duplicated
        # primitive creation
        _val = {}
        for k in set(six.iterkeys(val)) - in_obj:
            _val[k] = val[k]

        if obj.discriminator:
            self[obj.discriminator] = ctx['name']

        return _val

    def cleanup(self, val, ctx):
        """
        """
        if ctx['addp'] == True:
            for k, v in six.iteritems(val):
                self[k] = v
            ctx['addp'] = False
        elif ctx['addp_schema'] != None:
            obj = ctx['addp_schema']
            for k, v in six.iteritems(val):
                self[k] = ctx['factory'].produce(obj.additionalProperties, v)
            ctx['addp_schema'] = None

        return {}

    def __eq__(self, other):
        """ equality operater, 
        will skip checking when both value are None or no attribute.

        :param other: another model
        :type other: primitives.Model or dict
        """
        if other == None:
            return False

        for k, v in six.iteritems(self):
            if v != other.get(k, None):
                return False

        residual = set(other.keys()) - set(self.keys())
        for k in residual:
            if other[k] != None:
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

